function [TEM,NLoG] = NLoG_method(...
image,smallest_scale_parameter,largest_NP_diameter,NMS_threshold)
%NLoG_method Measures the particle size distribution of a TEM
%image using a series of Laplacian of Gaussian (LoG) filters and subsequent
%hysteresis thresholding.
%
%   image: the path to the TEM image with apostrophes, like:
%   'D:\TEM_measurements\TEM.dm3'
%   smallest_scale_parameter: the smallest scale parameter in the sampled
%   scale space, in pixels
%   largest_NP_diameter: the diameter of the largest nanoparticle (NP) in
%   the image, in nanometer
%
%	Example usage: 
%	[TEM,NLoG] = NLoG_method('D:\TEM_measurements\TEM.dm3',4,16);

% Pre-amble
if isempty(gcp('nocreate'))
    parpool('local');
end

% Import TEM image
TEM = import_TEM(image);

tic % Start timing the method

% Pre-process image (normalize, auto-contrast and invert if needed)
TEM = pre_process(TEM);

% Create scale-space
NLoG = create_scale_space(smallest_scale_parameter,largest_NP_diameter,TEM);

% Generate LoG filters
NLoG = create_NLoG(NLoG);

% Apply filters to TEM image
NLoG = apply_filters(TEM,NLoG);

% Apply Non-Maxima-Suppression (NMS)
NLoG = NMS(NLoG,NMS_threshold);

% Remove overlapping NPs and obtain data for automatically found NPs
NLoG = remove_overlapping_NPs(NLoG);

% Get NP sizes
NLoG.sizes = TEM.pixelsize .*...
    sort(sqrt(cell2mat(struct2cell(regionprops...
    (NLoG.L, 'Area'))') ./ pi));

% Plot the TEM image with encircled nanocrystals for verification
plot_verification(TEM,NLoG);

% Stop timing the method.
elapsedTime = toc %#ok<*NOPRT>
NLoG.time = elapsedTime; 

end

function [TEM] = import_TEM(image)
%IMPORT_TEM Imports TEM image
%   image: the path to the TEM image with apostrophes, like:
%   'D:\TEM_measurements\TEM.tif'

[~,name,EXT]=fileparts(image);
if strcmp(EXT,'.tif')
    temp = im2double(imread(image));
    TEM.data = temp(:,:,1);
    TEM.ext = EXT;
    TEM.size = input(' enter size of TEM image (in nm): \n');
    TEM.pixelsize = TEM.size/length(TEM.data);
elseif strcmp(EXT,'.TIF')
    temp = im2double(imread(image));
    TEM.data = temp(:,:,1);
    TEM.ext = EXT;
    TEM.size = input(' enter size of TEM image (in nm): \n');
    TEM.pixelsize = TEM.size/length(TEM.data);
elseif strcmp(EXT,'.jpg')
    TEM.data = im2double(imread(image));
    TEM.ext = EXT;
    TEM.size = input(' enter size of TEM image (in nm): \n');
    TEM.pixelsize = TEM.size/length(TEM.data);
elseif strcmp(EXT,'.bmp')
    TEM.data = im2double(imread(image));
    TEM.data = TEM.data(:,:,1);
    TEM.ext = EXT;
    TEM.size = input(' enter size of TEM image (in nm): \n');
    TEM.pixelsize = TEM.size/length(TEM.data);
end

TEM.original = TEM.data;
TEM.preprocessed = 'no';
TEM.sample = name;
TEM.font = 9;

% Create scalebar
TEM.scalebar.length_in_nm = 20;
TEM.scalebar.length_in_pix = TEM.scalebar.length_in_nm/TEM.pixelsize;
TEM.scalebar.origin = [0.05*length(TEM.data) 0.95*length(TEM.data)];
TEM.scalebar.end =  TEM.scalebar.origin + [TEM.scalebar.length_in_pix 0];

end

function [TEM] = pre_process(TEM)
%PRE-PROCESS image with normalization and auto-contrast

if strcmp(TEM.preprocessed,'no')
    % Normalize image
    TEM.data = TEM.data-min(TEM.data(:));
    TEM.data = TEM.data./max(TEM.data(:));
    
    % Change contrast
    hgram = normpdf(0:1/(511):1,0.5,0.22);
    [TEM.data,TEM.histeq_T] = histeq(TEM.data,hgram);
    
    % Plot image
    orig_fig = figure;
    plot_TEM(TEM)
    
    % Invert if needed
    prompt = 'For analysis purposes the nanoparticles should be black. Do you want to invert the image? y/n [n]: ';
    str = input(prompt,'s');
    
    if strcmp(str,'y')
        TEM.original = TEM.data;
        TEM.data = imcomplement(TEM.data);
    end
    
    plot_TEM(TEM)
    pause(2)
    close(orig_fig)
    
    TEM.preprocessed = 'yes';
end

end

function [auto] = ...
    create_scale_space(smallest_scale_parameter,largest_NP_diameter,TEM)
%CREATE_SCALE_SPACE Creates a sampled scale space

auto.scale_space = ...
    (smallest_scale_parameter:1:((0.5*largest_NP_diameter)/TEM.pixelsize));

end

function [auto] = create_NLoG(auto)
%CREATE_NLoG generates LoG filters and normalizes them

scale_space = auto.scale_space;
parfor i = 1:length(scale_space)
    hsize = round(7*scale_space(i));
    LoG{i}.data = single(scale_space(i)^2 .*...
        fspecial('log',hsize,scale_space(i)));
end
auto.LoG = LoG;

end

function [auto] = apply_filters(TEM,auto)
%APPLY_FILTERS applies LoG filters to a TEM image by convolving them.

temp = single(TEM.data);
LoG = auto.LoG;
temp2 = conv2(temp, LoG{1}.data, 'same');
results = single(zeros(length(temp2),length(temp2),length(LoG)));
results(:,:,1) = temp2;
clear temp2

parfor i = 2:length(LoG)
    results(:,:,i) = conv2(temp, LoG{i}.data, 'same');
end
auto.results = results;

end

function [auto] = NMS(auto,NMS_thresh)
%NMS Non-Maxima-Suppression
%   This function applies non-maxima-suppression (NMS) to the 3D result
%   (x,y,scale-space) of LoG filters convolutions.

results = padarray(auto.results,[2 2 2]);
msksize = 5;
siz = size(auto.results);
msk = true(msksize,msksize,msksize);
msk(round(msksize/2),round(msksize/2),round(msksize/2)) = false;
results_dil = imdilate(results,msk);
NPpos3D = results > results_dil & results>NMS_thresh;

NPpos3D = circshift(NPpos3D,[-2 -2 -2]);
auto.NPpos3D = NPpos3D(1:siz(1),1:siz(2),1:length(auto.scale_space));
auto.NPpos2D = sum(NPpos3D,3);
end

function [auto] = remove_overlapping_NPs(auto)
%REMOVE_OVERLAPPING_NPS Removes overlapping nanoparticles detected with the
%NLoG method.

NPpos3D = auto.NPpos3D;
NPpos2D = auto.NPpos2D;

NPpos2D_r_in_pix = single(zeros(size(NPpos2D)));
NP_auto = single(zeros(size(NPpos2D)));

[columnsInImage, rowsInImage] = meshgrid(1:length(NP_auto),...
    1:length(NP_auto));
scale_space = auto.scale_space;
results = auto.results;
[row,col]=find(NPpos2D>0);
for i=1:length(row)
    ind = find(NPpos3D(row(i),col(i),:)==1);
    if length(ind) > 1
        [~,ind2] = max(results(row(i),col(i),ind));
        ind = ind(ind2);
    end
    NPpos2D_r_in_pix(row(i),col(i)) = scale_space(ind);
    radius = 1.5222*NPpos2D_r_in_pix(row(i),col(i));
    temp = (rowsInImage - row(i)).^2 ...
        + (columnsInImage - col(i)).^2 <= radius.^2;
    NP_auto = NP_auto + temp;
end

NP_auto(NP_auto>1)=1;
auto.data = NP_auto;
[auto.B,auto.L,auto.N] = bwboundaries(auto.data,8,'noholes');

end

function plot_TEM(TEM)
%PLOT_TEM Plots the TEM image.

imagesc(TEM.data)
colormap(gray)
axis equal
xlim([0 length(TEM.data)])
ylim([0 length(TEM.data)])
axis off
hold on
plot([TEM.scalebar.origin(1),TEM.scalebar.end(1)],...
    [TEM.scalebar.origin(2),TEM.scalebar.end(2)],'Color','w','LineWidth',3)
scalebar_text = text(TEM.scalebar.origin(1),0.9*length(TEM.data),...
    sprintf('%0.0f nm',TEM.scalebar.length_in_nm));
set(scalebar_text,'FontSize',TEM.font,'Color','w')
end

function plot_verification(TEM,auto)
%PLOT_VERIFICATION plots the TEM image and encircled the detected
%nanoparticles with a green line.

plot_TEM(TEM)
for k = 1:length(auto.B)
    boundary = auto.B{k};
    plot(boundary(:,2), boundary(:,1), 'g', 'LineWidth', 1)
end
hold off

end